import numpy as np

# don't import any random modules in a Pythran file. Here, no problem!
from fluiddyn.util import mpi

from transonic import boost

# TRANSONIC_NO_IMPORT

# transonic import numpy as np

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


@boost
def func(a, b):
    return (a * np.log(b)).max()


if __name__ == "__main__":

    n0, n1 = 100, 200
    a0 = np.random.rand(n0, n1)
    a1 = np.random.rand(n0, n1)

    result = func(a0, a1)
    if mpi.nb_proc > 1:
        result = mpi.comm.allreduce(result, op=mpi.MPI.MAX)
    mpi.printby0(result)

    a0 = (1000 * a0).astype(int)
    result = func(a0, a1)
    if mpi.nb_proc > 1:
        result = mpi.comm.allreduce(result, op=mpi.MPI.MAX)
    mpi.printby0(result)



import numpy as np

# pythran export __for_method__Transmitter____call__(float, float[:])
# pythran export __for_method__Transmitter____call__(float, int[:])
# pythran export __for_method__Transmitter____call__(float, float[:, :])
# pythran export __for_method__Transmitter____call__(float, int[:, :])


def __for_method__Transmitter____call__(self_freq, inp):
    "My docstring"
    print("__for_method...")
    return inp * np.exp(((np.arange(len(inp)) * self_freq) * 1j))


# pythran export __code_new_method__Transmitter____call__

__code_new_method__Transmitter____call__ = """

def new_method(self, inp):
    print("new_method")
    return pythran_func(self.freq, inp)

"""



"""Create Pythran signatures from type hints
============================================

User API
--------

.. autoclass:: Type
   :members:
   :private-members:

.. autoclass:: NDim
   :members:
   :private-members:

.. autoclass:: Shape
   :members:
   :private-members:

.. autoclass:: Array
   :members:
   :private-members:

.. autoclass:: Union
   :members:
   :private-members:

Internal API
------------

.. autoclass:: TemplateVar
   :members:
   :private-members:

.. autoclass:: ArrayMeta
   :members:
   :private-members:

.. autofunction:: compute_pythran_types_from_types

.. autofunction:: compute_pythran_types_from_valued_types

.. autofunction:: make_signature_from_template_variables

.. autofunction:: make_signatures_from_typehinted_func

"""

import itertools
import inspect


class TemplateVar:
    """Base class for template variables

    >>> T = TemplateVar("T")
    >>> T = TemplateVar("T", int, float)

    >>> T = TemplateVar()  # raise ValueError
    >>> T = TemplateVar(1)  # raise TypeError

    """

    _type_values = type
    _letter = "T"

    def get_template_parameters(self):
        return (self,)

    def __init__(self, *args, _ts=None):

        if not args:
            raise ValueError

        if _ts is None:
            ts = _get_transonic_calling_module()
        else:
            ts = _ts

        if type(self) not in ts.names_template_variables:
            ts.names_template_variables[type(self)] = set()

        names_already_used = ts.names_template_variables[type(self)]

        if self._is_correct_for_name(args[0]):
            self.__name__ = args[0]
            args = args[1:]
        else:
            index_var = len(names_already_used)

            while self._letter + str(index_var) in names_already_used:
                index_var += 1

            self.__name__ = self._letter + str(index_var)

        self.values = args

        names_already_used.add(self.__name__)

        self._check_type_values()

    def _is_correct_for_name(self, arg):
        return isinstance(arg, str)

    def _check_type_values(self):
        if not all(isinstance(value, self._type_values) for value in self.values):
            raise TypeError(
                f"{self.values} "
                f"{[isinstance(value, self._type_values) for value in self.values]}"
            )


class Type(TemplateVar):
    """Template variable representing the dtype of an array"""

    def __repr__(self):
        return self.__name__

    def get_pythran_type(self, **kwargs):

        dtype = None

        for key, value in kwargs.items():
            if key == self.__name__:
                dtype = value
                break

        if dtype is None:
            raise ValueError

        return dtype.__name__


class NDim(TemplateVar):
    """Template variable representing the number of dimension of an array"""

    _type_values = int
    _letter = "N"

    def __init__(self, *args, shift=0, _ts=None):

        if _ts is None:
            _ts = _get_transonic_calling_module()

        super().__init__(*args, _ts=_ts)
        self.shift = shift

    def __repr__(self):

        name = self.__name__

        if self.shift < 0:
            name = name + f" - {abs(self.shift)}"
        elif self.shift > 0:
            name = name + f" + {abs(self.shift)}"

        return name

    def __add__(self, number):
        ts = _get_transonic_calling_module()
        return type(self)(self.__name__, *self.values, shift=number, _ts=ts)

    def __sub__(self, number):
        ts = _get_transonic_calling_module()
        return type(self)(self.__name__, *self.values, shift=-number, _ts=ts)


class Shape(TemplateVar):
    """Shape template variable

    NotImplemented!

    """

    _letter = "S"
    _type_values = str, list, tuple

    def _is_correct_for_name(self, arg):
        raise NotImplementedError
        return isinstance(arg, str)


class UnionVar(TemplateVar):

    _type_values = type
    _letter = "U"


class ArrayMeta(type):
    """Metaclass for the Array class used for type hints"""

    def __getitem__(self, parameters):

        if not isinstance(parameters, tuple):
            parameters = (parameters,)

        dtype = None
        ndim = None
        params_filtered = []
        for param in parameters:
            if isinstance(param, (Type, type)):
                if dtype is not None:
                    raise ValueError(
                        "Array should be defined with only one variable defining "
                        "the types. For more than one type, use "
                        "for example Type(float, int)"
                    )
                dtype = param

            if isinstance(param, NDim):
                if ndim is not None:
                    raise ValueError(
                        "Array should be defined with only "
                        "one NDim. For more than one dimension, use "
                        "for example NDim(2, 3)."
                    )
                ndim = param

            if (
                isinstance(param, str)
                and param[-1] == "d"
                and param[:-1].isnumeric()
            ):
                try:
                    tmp = int(param[:-1])
                except ValueError:
                    pass
                else:
                    if ndim is not None:
                        raise ValueError(
                            "Array should be defined with only "
                            "one string fixing the number of dimension. "
                            "Use for example NDim(2, 3)."
                        )
                    param = ndim = NDim(tmp, _ts=_get_transonic_calling_module())

            if isinstance(param, str):
                raise ValueError(f"{param} cannot be interpretted...")

            params_filtered.append(param)

        if dtype is None:
            raise ValueError("No way to determine the dtype of the array")

        if ndim is None:
            raise ValueError("No way to determine the ndim of the array")

        parameters = {p.__name__: p for p in params_filtered}

        ArrayBis = type(
            "ArrayBis",
            (Array,),
            {"dtype": dtype, "ndim": ndim, "parameters": parameters},
        )

        return ArrayBis

    def get_parameters(self):
        return getattr(self, "parameters", dict())

    def get_template_parameters(self):
        return tuple(
            param
            for param in self.get_parameters().values()
            if isinstance(param, TemplateVar)
        )

    def __repr__(self):
        strings = []
        for p in self.parameters.values():
            if isinstance(p, type):
                string = p.__name__
            else:
                string = repr(p)
            strings.append(string)

        return "Array[" + ", ".join(strings) + "]"

    def get_pythran_type(self, **kwargs):

        dtype = ndim = None

        for var in self.parameters.values():
            if isinstance(var, Type) and var.values:
                dtype = var.values[0]
            elif isinstance(var, NDim) and var.values:
                ndim = var.values[0]
            elif isinstance(var, type):
                dtype = var

        for key, value in kwargs.items():
            try:
                template_var = self.parameters[key]
            except KeyError:
                continue

            if isinstance(template_var, Type):
                dtype = value
            elif isinstance(template_var, NDim):
                ndim = value + template_var.shift
            else:
                raise ValueError

        if dtype is None or ndim is None:
            raise ValueError

        base = f"{dtype.__name__}"
        if ndim == 0:
            return base
        return base + f"[{', '.join([':']*ndim)}]"


class Array(metaclass=ArrayMeta):
    """Represent a Numpy array in type hints"""

    pass


class UnionMeta(type):
    """Metaclass for the Array class used for type hints"""

    def __getitem__(self, types):

        if not isinstance(types, tuple):
            types = (types,)

        ts = _get_transonic_calling_module()
        template_var = UnionVar(*types, _ts=ts)

        return type(
            "UnionBis", (Union,), {"types": types, "template_var": template_var}
        )

    def get_template_parameters(self):
        template_params = []
        for type_ in self.types:
            if hasattr(type_, "get_template_parameters"):
                template_params.extend(type_.get_template_parameters())

        template_params.append(self.template_var)

        return tuple(template_params)

    def __repr__(self):
        strings = []
        for p in self.parameters.values():
            if isinstance(p, type):
                string = p.__name__
            else:
                string = repr(p)
            strings.append(string)
        return "Union[" + ", ".join(strings) + "]"

    def get_pythran_type(self, **kwargs):
        type_ = kwargs.pop(self.template_var.__name__)
        return compute_pythran_type_from_type(type_, **kwargs)


class Union(metaclass=UnionMeta):
    """Similar to typing.Union

    >>> U = Union[float, Array[int, "1d"]]

    """

    pass


normalized_types = {"float": "float64", "complex": "complex128"}


def normalize_type_name(name):
    try:
        return normalized_types[name]
    except KeyError:
        return name


def compute_pythran_type_from_type(type_, **kwargs):
    if hasattr(type_, "get_pythran_type"):
        pythran_type = type_.get_pythran_type(**kwargs)
    elif hasattr(type_, "__name__"):
        pythran_type = type_.__name__
    else:
        pythran_type = str(type_)
        types = pythran_type.split(" or ")
        new_types = []
        for _type in types:
            _type = normalize_type_name(_type)
            if "][" in _type:
                # C style: we try to rewrite it in Cython style
                base, dims = _type.split("[", 1)
                dims = ", ".join([_ or ":" for _ in dims[:-1].split("][")])
                _type = normalize_type_name(base) + "[" + dims + "]"
            elif _type.endswith("[]"):
                _type = normalize_type_name(_type[:-2]) + "[:]"
            new_types.append(_type)
        pythran_type = " or ".join(new_types)

    return normalize_type_name(pythran_type)


def compute_pythran_types_from_types(types, **kwargs):
    """Compute a list of pythran types

    """
    pythran_types = []
    for type_ in types:
        pythran_types.append(compute_pythran_type_from_type(type_, **kwargs))

        if "_empty" in pythran_types:
            raise ValueError(
                "At least one annotation type lacking in a signature.\n"
                f"types = {types}"
            )

    return pythran_types


def compute_pythran_types_from_valued_types(types):
    """Compute a list of pythran types

    """
    template_parameters = []

    for type_ in types:
        if hasattr(type_, "get_template_parameters"):
            template_parameters.extend(type_.get_template_parameters())

    template_parameters = set(template_parameters)

    if not template_parameters:
        str_types = []
        for type_ in types:
            if isinstance(type_, type):
                str_type = type_.__name__
            else:
                str_type = type_
            str_types.append(str_type)

        if "_empty" in str_types:
            raise ValueError(
                "At least one annotation type lacking in a signature.\n"
                f"types = {types}"
            )

        return (str_types,)

    if not all(param.values for param in template_parameters):
        raise ValueError

    values_template_parameters = {}
    for param in template_parameters:
        values_template_parameters[param.__name__] = param.values

    pythran_types = []
    names = values_template_parameters.keys()
    for set_types in itertools.product(*values_template_parameters.values()):
        template_variables = {
            name: value for name, value in zip(names, set_types)
        }

        pythran_types.append(
            compute_pythran_types_from_types(types, **template_variables)
        )

    return pythran_types


def make_signature_from_template_variables(func, _signature=None, **kwargs):
    """Create signature for a function with values for the template types

    Parameters
    ----------

    func: a function

    kwargs : dict

        The template types and their value

    """
    if _signature is None:
        _signature = inspect.signature(func)

    types = [param.annotation for param in _signature.parameters.values()]

    pythran_types = compute_pythran_types_from_types(types, **kwargs)

    # "multiply" the signatures to take into account the "or" syntax
    multi_pythran_types = [
        _ for _ in itertools.product(*[t.split(" or ") for t in pythran_types])
    ]
    signatures = []
    for pythran_types in multi_pythran_types:
        signatures.append(f"{func.__name__}(" + ", ".join(pythran_types) + ")")

    return signatures


def make_signatures_from_typehinted_func(func):
    """Make the signatures from annotations if it is possible

    Useful when there are only "not templated" types.

    """
    annotations = func.__annotations__

    if not annotations:
        return tuple()

    types = annotations.values()

    template_parameters = []

    for type_ in types:
        if hasattr(type_, "get_template_parameters"):
            template_parameters.extend(type_.get_template_parameters())

    template_parameters = set(template_parameters)

    _signature = inspect.signature(func)

    if not template_parameters:
        signatures = make_signature_from_template_variables(
            func, _signature=_signature
        )
        return signatures

    if not all(param.values for param in template_parameters):
        return tuple()

    values_template_parameters = {}
    for param in template_parameters:
        values_template_parameters[param.__name__] = param.values

    names = values_template_parameters.keys()
    signatures = []
    for set_types in itertools.product(*values_template_parameters.values()):
        template_variables = {
            name: value for name, value in zip(names, set_types)
        }
        signatures.extend(
            make_signature_from_template_variables(
                func, _signature=_signature, **template_variables
            )
        )

    return signatures


# we need to put this import here
from .aheadoftime import _get_transonic_calling_module




import numpy as np
# transonic import numpy as np


class Transmitter():

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp))*self.freq*1j)



import numpy as np

# transonic import numpy as np

from transonic import jit, wait_for_all_extensions

import numba


def laplace_numpy(image):
    """Laplace operator in NumPy for 2D images."""
    laplacian = (
        image[:-2, 1:-1] + image[2:, 1:-1] + image[1:-1, :-2] + image[1:-1, 2:]
        - 4*image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh


@jit
def laplace_pythran(image):
    """Laplace operator in NumPy for 2D images."""
    laplacian = (
        image[:-2, 1:-1] + image[2:, 1:-1] + image[1:-1, :-2] + image[1:-1, 2:]
        - 4*image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh


@jit
def laplace_pythran_loops(image):
    """Laplace operator for 2D images."""
    h = image.shape[0]
    w = image.shape[1]
    laplacian = np.empty((h - 2, w - 2))
    for i in range(1, h - 1):
        for j in range(1, w - 1):
            laplacian[i-1, j-1] = (
                np.abs(image[i-1, j] + image[i+1, j] + image[i, j-1]
                + image[i, j+1] - 4*image[i, j]) > 0.05
            )
    return laplacian


@numba.jit(nopython=True, cache=True)
def laplace_numba(image):
    """Laplace operator in NumPy for 2D images. Numba accelerated."""
    laplacian = (
        image[:-2, 1:-1] + image[2:, 1:-1] + image[1:-1, :-2] + image[1:-1, 2:]
        - 4*image[1:-1, 1:-1]
    )
    thresh = np.abs(laplacian) > 0.05
    return thresh


@numba.jit(nopython=True, cache=True)
def laplace_numba_loops(image):
    """Laplace operator for 2D images. Numba accelerated."""
    h = image.shape[0]
    w = image.shape[1]
    laplacian = np.empty((h - 2, w - 2))
    for i in range(1, h - 1):
        for j in range(1, w - 1):
            laplacian[i-1, j-1] = (
                np.abs(image[i-1, j] + image[i+1, j] + image[i, j-1]
                + image[i, j+1] - 4*image[i, j]) > 0.05
            )
    return laplacian


if __name__ == "__main__":
    from skimage.data import astronaut
    from skimage.color import rgb2gray

    image = astronaut()
    image = rgb2gray(image)

    laplace_pythran(image)
    laplace_pythran_loops(image)

    laplace_numba(image)
    laplace_numba_loops(image)

    wait_for_all_extensions()


"""
With
- clang 6.0
- pythran 0.9
- numba 0.40.1

%timeit laplace_numpy(image)
1.92 ms ± 69.7 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

%timeit laplace_pythran(image)
211 µs ± 56.2 ns per loop (mean ± std. dev. of 7 runs, 1000 loops each)

%timeit laplace_pythran_loops(image)
241 µs ± 821 ns per loop (mean ± std. dev. of 7 runs, 1000 loops each)

%timeit laplace_numba(image)
2.05 ms ± 1.09 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)

%timeit laplace_numba_loops(image)
240 µs ± 489 ns per loop (mean ± std. dev. of 7 runs, 1000 loops each)

"""



from tokenize import tokenize, untokenize, NAME, OP
from token import tok_name
from io import BytesIO
import inspect

import black

import numpy as np

from transonic.log import logger
from transonic.util import get_source_without_decorator
from transonic.annotation import compute_pythran_types_from_valued_types


def trans_def_method(func):
    func.__transonic__ = "trans_def_method"
    return func


# logger.set_level("debug")


from transonic import Array, Type, NDim

A = Array[Type(float, int), NDim(1, 2)]


class Transmitter:
    freq: float

    def __init__(self):
        pass

    @trans_def_method
    def __call__(self, inp: A):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)

    @trans_def_method
    def call_with_print(self, inp: A):
        """call + print"""
        print("call_with_print")
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)


def produce_pythran_code_class(cls):

    pythran_code = ""

    for key, value in cls.__dict__.items():
        if (
            hasattr(value, "__transonic__")
            and value.__transonic__ == "trans_def_method"
        ):
            pythran_code += produce_pythran_code_class_func(cls, key)

    return pythran_code


def produce_pythran_code_class_func(cls, func_name):

    cls_name = cls.__name__
    cls_annotations = cls.__annotations__

    func = cls.__dict__[func_name]

    signature = inspect.signature(func)
    types_func = [param.annotation for param in signature.parameters.values()][1:]
    args_func = list(signature.parameters.keys())[1:]

    src = get_source_without_decorator(func)

    tokens = []
    attributes = set()

    using_self = False

    g = tokenize(BytesIO(src.encode("utf-8")).readline)
    for toknum, tokval, a, b, c in g:
        logger.debug((tok_name[toknum], tokval))

        if using_self == "self":
            if toknum == OP and tokval == ".":
                using_self = tokval
                continue
            elif toknum == OP and tokval == ",":
                tokens.append((NAME, "self"))
                using_self = False
            else:
                raise RuntimeError

        if using_self == ".":
            if toknum == NAME and tokval in cls_annotations:
                using_self = False
                tokens.append((NAME, "self_" + tokval))
                attributes.add(tokval)
                continue
            else:
                raise RuntimeError

        if toknum == NAME and tokval == "self":
            using_self = tokval
            continue

        tokens.append((toknum, tokval))

    attributes = sorted(attributes)

    types_attrs = [cls_annotations[attr] for attr in attributes]

    attributes_self = ["self_" + attr for attr in attributes]

    index_self = tokens.index((NAME, "self"))

    tokens_attr = []
    for ind, attr in enumerate(attributes_self):
        tokens_attr.append((NAME, attr))
        tokens_attr.append((OP, ","))

    tokens = tokens[:index_self] + tokens_attr + tokens[index_self + 2 :]

    index_func_name = tokens.index((NAME, func_name))
    name_pythran_func = f"__for_method__{cls_name}__{func_name}"
    tokens[index_func_name] = (NAME, name_pythran_func)

    new_code = untokenize(tokens).decode("utf-8")
    new_code = black.format_str(new_code, line_length=82)

    # args_pythran = attributes + args_func
    types_pythran = types_attrs + types_func

    pythran_signatures = "\n"

    for types_string_signature in compute_pythran_types_from_valued_types(
        types_pythran
    ):
        pythran_signatures += (
            "# pythran export "
            + name_pythran_func
            + "("
            + ", ".join(types_string_signature)
            + ")\n"
        )

    pythran_code = pythran_signatures + "\n" + new_code

    name_var_code_new_method = f"__code_new_method__{cls_name}__{func_name}"

    str_self_dot_attributes = ", ".join("self." + attr for attr in attributes)
    str_args_func = ", ".join(args_func)

    pythran_code += (
        f"\n# pythran export {name_var_code_new_method}\n"
        f'\n{name_var_code_new_method} = """\n\n'
        f"def new_method(self, {str_args_func}):\n"
        f"    return pythran_func({str_self_dot_attributes}, {str_args_func})"
        '\n\n"""\n'
    )

    return pythran_code


pythran_code = produce_pythran_code_class(Transmitter)
print(pythran_code)



import numpy as np

# don't import skimage in a Pythran file. Here, no problem!
from skimage.filters import sobel

from transonic import boost

# transonic import numpy as np

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


@boost
def func(a: float, b: float):
    return (a * np.log(b)).max()


@boost
def func1(a: int, b: float):
    return a * np.cos(b)



import numpy as np

from transonic import boost, include

# transonic import numpy as np

# transonic def func(float[][], float[][])
# transonic def func(int[][], float[][])


@include
def my_log(b):
    return np.log(b)


@boost
def func(a, b):
    return (a * my_log(b)).max()



from transonic import jit, Type

T = Type(int, float)

@jit()
def func(a: T, b: T):
    return a * b

if __name__ == "__main__":

    from time import sleep

    a_i = b_i = 1
    a_f = b_f = 1.

    for _ in range(10):
        print(_, end=",", flush=True)
        func(a_i, b_i)
        sleep(1)

    print()

    for _ in range(10):
        print(_, end=",", flush=True)
        func(a_f, b_f)
        sleep(1)



import numpy as np
import transonic as ts
from transonic import Type, NDim, Array

T = Type("T")
N = NDim("N")

A = Array[T, N]
A1 = Array[np.int8, N + 1]


@ts.boost
def compute(a: A, b: A, c: T, d: A1, e: str):
    print(e)
    tmp = a + b
    return tmp


for dtype in [int, np.complex128]:
    for ndim in [1, 3]:
        ts.make_signature(compute, T=dtype, N=ndim)


if __name__ == "__main__":
    print(__name__)

    shape = [2]

    a = b = np.zeros(shape, dtype=int)
    d = np.zeros(shape + [2], dtype=np.int8)

    compute(a, b, 5, d, "hello")



"""
See https://bitbucket.org/fluiddyn/fluidpythran/issues/29/

"""
from transonic import boost, with_blocks, block


def non_pythranizable(arg):
    """represent a function that can not be compiled by Pythran"""
    return arg


@boost
class MyClass:

    attr0: int

    @with_blocks
    def func(self, arg: int):

        a = non_pythranizable(arg)

        with block(a=float):
            tmp = a + arg + self.attr0

        return non_pythranizable(tmp)



import shutil
from distutils.core import Distribution
from copy import copy
import pytest

from . import dist
from .mpi import nb_proc

from .dist import (
    detect_pythran_extensions,
    modification_date,
    make_backend_files,
    ParallelBuildExt,
)
from . import path_data_tests


can_actually_import_pythran = copy(dist.can_import_pythran)


def setup_module():
    dist.can_import_pythran = True


@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_detect_pythran_extensions():

    shutil.rmtree(path_data_tests / "__pythran__", ignore_errors=True)

    names = [
        "block_fluidsim.py",
        "class_blocks.py",
        "no_pythran_.py",
        "type_hint_notemplate.py",
        "blocks_type_hints.py",
        "classic.py",
        "mixed_classic_type_hint.py",
        "type_hint.py",
    ]

    make_backend_files(
        (path_data_tests / name for name in names),
        mocked_modules=("toto.titi", "numpy"),
    )
    ext_names = detect_pythran_extensions(path_data_tests)
    assert len(ext_names) == len(names) - 1

    shutil.rmtree(path_data_tests / "__pythran__", ignore_errors=True)


@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_modification_date():

    modification_date(path_data_tests / "no_pythran_.py")


@pytest.mark.skipif(nb_proc > 1, reason="No dist in MPI")
def test_build_ext():
    dist = Distribution()
    build_ext = ParallelBuildExt(dist)

    build_ext.initialize_options()
    build_ext.parallel = 1
    build_ext.finalize_options()


def teardown_module():
    dist.can_import_pythran = can_actually_import_pythran



"""Command line
===============

Internal API
------------

.. autofunction:: main

"""

import subprocess
import sys
import logging
from pathlib import Path
import sysconfig
from time import time, sleep

logger = logging.getLogger("pythran_fluid")
logger.setLevel(logging.INFO)

ext_suffix = sysconfig.get_config_var("EXT_SUFFIX") or ".so"


def main():
    """Minimal layer above the Pythran commandline"""

    assert sys.argv[0].endswith("transonic_cl/run_pythran.py")

    args = sys.argv[1:]

    name = args[0]
    path = Path.cwd() / name

    if "-o" in args:
        index_output = args.index("-o") + 1
        name_out = args[index_output]
    else:
        name_out = Path(name).with_suffix(ext_suffix).name

    name_out_base = name_out.split(".", 1)[0]

    if "-o" in args:
        name_tmp = name_out_base + ".tmp"
        args[index_output] = name_tmp
        path_tmp = Path(name_tmp)
        path_out = path_tmp.with_suffix(ext_suffix)
    else:
        path_out = Path(name_out)

    name_lock = Path(name_out_base + ".lock")

    if name_lock.exists():
        print(
            f"lock file {name_lock.absolute()} present: "
            "waiting for completion of the compilation",
            flush=True,
        )
        time_out_lock = 3600  # (s) let's hope it's enough
        time_start = time()
        while name_lock.exists() and time() - time_start < time_out_lock:
            sleep(1)
        if time() - time_start >= time_out_lock:
            logger.error(f"Remove lock file {name_lock.absolute()}")
            name_lock.unlink()
            raise TimeoutError(
                f"Stop waiting for a lock file to be deleted {name_lock.absolute()}"
            )

        assert not name_lock.exists()
        sleep(1)
        if not path_out.exists():
            raise RuntimeError(
                f"After lock file were deleted, {path_out.absolute()} not created"
            )

        return

    if "-v" in args:
        # no capture_output
        stdout = stderr = None
    else:
        stdout = stderr = subprocess.PIPE

    print(f"Pythranizing {path}", flush=True)
    args.insert(0, "pythran")
    name_lock.touch()
    try:
        completed_process = subprocess.run(args, stdout=stdout, stderr=stderr)
    except Exception:
        pass
    finally:
        name_lock.unlink()

    if "-o" in args and path_tmp.exists():
        path_tmp.rename(path_out)

    if path_out.exists():
        print(f"File {path_out.absolute()} created by Pythran")
    else:
        logger.error(
            f"Error! File {path_out.absolute()} has not been created by Pythran"
        )
        try:
            completed_process
        except NameError:
            pass
        else:
            if completed_process.stdout:
                print(f"Pythran stdout:\n{completed_process.stdout}")
            if completed_process.stderr:
                logger.error(f"Pythran stderr:\n{completed_process.stderr}")


if __name__ == "__main__":
    main()



import numpy as np
# transonic import numpy as np

from transonic import jit, include

@include("func1")
def func0(a, b):
    return a + b

@jit()
def func1(a, b):
    return np.exp(a) * b * func0(a, b)

if __name__ == "__main__":

    from time import sleep

    a = b = np.zeros([2, 3])

    for i in range(40):
        print(f"{i}, call with arrays", end=", ")
        func1(a, b)
        print("call with numbers")
        func1(1, 1.)
        sleep(1)



"""Internal utilities
=====================

Public API
----------

.. autofunction:: set_compile_at_import

Internal API
------------

.. autofunction:: find_module_name_from_path

.. autofunction:: get_module_name

.. autofunction:: modification_date

.. autofunction:: has_to_build

.. autofunction:: get_source_without_decorator

.. autoclass:: TypeHintRemover
   :members:
   :private-members:

.. autofunction:: strip_typehints

.. autofunction:: get_ipython_input

.. autofunction:: get_info_from_ipython

.. autofunction:: has_to_compile_at_import

.. autofunction:: import_from_path

.. autofunction:: query_yes_no

.. autofunction:: clear_cached_extensions

.. autofunction:: is_method
"""

import os
import sys
import inspect
import re
from pathlib import Path
import ast
import importlib.util
from distutils.util import strtobool
import shutil
from textwrap import dedent

from typing import Callable

import astunparse

try:
    import black
except ImportError:
    black = False

try:
    import pythran
except ImportError:
    pythran = False

try:
    from IPython.core.getipython import get_ipython
except ImportError:
    pass

from .pythranizer import (
    ext_suffix,
    name_ext_from_path_backend,
    make_hex,
    modification_date,
    has_to_build,
)

from .config import path_root
from .compat import fspath

__all__ = ["modification_date", "has_to_build"]


path_jit_classes = path_root / "__jit_classes__"


def find_module_name_from_path(path_py: Path):
    """Find the module name from the path of a Python file

    It is done by looking to ``sys.path`` to see how the module can be imported.

    """

    cwd = Path.cwd()
    path = path_py.absolute().parent
    module_name = path_py.stem

    # special case for jit_classes
    try:
        path_rel = path.relative_to(path_jit_classes)
    except ValueError:
        pass
    else:
        tmp = [path_jit_classes.name]
        name_pack = fspath(path_rel).replace(os.path.sep, ".")
        if name_pack:
            tmp.append(name_pack)
        tmp.append(module_name)
        return ".".join(tmp)

    while path.parents:
        if path == cwd or fspath(path) in sys.path:
            return module_name

        module_name = path.name + "." + module_name
        path = path.parent

    return path_py.stem


def get_module_name(frame):
    """Get the full module name"""
    module = inspect.getmodule(frame[0])
    filename = None
    if module is not None:
        module_name = module.__name__
        if module_name in ("__main__", "<run_path>"):
            filename = frame.filename
    else:
        filename = frame.filename

    if filename is not None:
        module_name = find_module_name_from_path(Path(filename))

    if module_name is None:
        # ipython ?
        src, module_name = get_info_from_ipython()

    return module_name


def get_source_without_decorator(func: Callable):
    """Get the source of a function without its decorator"""
    src = inspect.getsource(func)
    src = dedent(src)
    return strip_typehints(re.sub(r"@.*?\sdef\s", "def ", src))


class TypeHintRemover(ast.NodeTransformer):
    """Strip the type hints

    from https://stackoverflow.com/a/42734810/1779806
    """

    def visit_FunctionDef(self, node):
        # remove the return type defintion
        node.returns = None
        # remove all argument annotations
        if node.args.args:
            for arg in node.args.args:
                arg.annotation = None
        return node


def strip_typehints(source):
    """Strip the type hints from a function"""

    if black:
        source = black.format_str(source, line_length=82)

    # parse the source code into an AST
    parsed_source = ast.parse(source)
    # remove all type annotations, function return type definitions
    # and import statements from 'typing'
    transformed = TypeHintRemover().visit(parsed_source)
    # convert the AST back to source code
    striped_code = astunparse.unparse(transformed)

    # bad hack to conserve the comments...
    sep = ":(\n)+    "
    head = re.split(sep, striped_code, maxsplit=1)[0]
    body = re.split(sep, source, maxsplit=1)[-1]
    return head + ":\n    " + body


def get_ipython_input(last=True):
    """Get the input code when called from IPython"""
    ip = get_ipython()

    hist_raw = ip.history_manager.input_hist_raw
    if last:
        return hist_raw[-1]
    else:
        return "\n".join(hist_raw)


def get_info_from_ipython():
    """Get the input code and a "filename" when called from IPython"""
    src = get_ipython_input()
    hex_input = make_hex(src)
    dummy_filename = "__ipython__" + hex_input
    return src, dummy_filename


_PYTHRANIZE_AT_IMPORT = None


def set_compile_at_import(value=True):
    """Control the "compile_at_import" mode"""
    global _PYTHRANIZE_AT_IMPORT
    _PYTHRANIZE_AT_IMPORT = value


def has_to_compile_at_import():
    """Check if transonic has to pythranize at import time"""
    if _PYTHRANIZE_AT_IMPORT is not None:
        return _PYTHRANIZE_AT_IMPORT
    return "TRANSONIC_COMPILE_AT_IMPORT" in os.environ


def import_from_path(path: Path, module_name: str):
    """Import a .py file or an extension from its path

    """
    if not path.exists():
        raise ImportError(
            f"File {path} does not exist. "
            f"[path.name for path in path.parent.glob('*')]:\n{[path.name for path in path.parent.glob('*')]}\n"
        )

    if "." in module_name:
        package_name, mod_name = module_name.rsplit(".", 1)
        name_file = path.name.split(".", 1)[0]
        if mod_name != name_file:
            module_name = ".".join((package_name, name_file))
    else:
        module_name = path.stem

    if module_name in sys.modules:
        module = sys.modules[module_name]

        if module.__file__.endswith(ext_suffix) and Path(module.__file__) == path:
            return module

    spec = importlib.util.spec_from_file_location(module_name, fspath(path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def query_yes_no(question: str, default: str = None, force: bool = False):
    """User yes or no query"""
    if force:
        return True

    if default is None:
        end = "(y/n)"
        default = ""
    elif default == "y":
        end = "([y]/n)"
    elif default == "n":
        end = "(y/[n])"

    print(f"{question} {end}")
    while True:
        answer = input()
        if answer == "":
            answer = default
        try:
            return strtobool(answer)
        except ValueError:
            print('Please respond with "y" or "n".')


def clear_cached_extensions(module_name: str, force: bool = False):
    """Delete the cached extensions related to a module

    """

    from .justintime import path_jit

    if module_name.endswith(".py"):
        module_name = module_name[:-3]

    if os.path.sep not in module_name:
        relative_path = module_name.replace(".", os.path.sep)
    else:
        relative_path = module_name

    path_pythran_dir_jit = path_jit / relative_path

    relative_path = Path(relative_path)

    path_pythran = relative_path.parent / (
        "__pythran__/" + relative_path.name + ".py"
    )
    path_ext = path_pythran.with_name(name_ext_from_path_backend(path_pythran))

    if not path_pythran_dir_jit.exists() and not (
        path_pythran.exists() or path_ext.exists()
    ):
        print(
            f"Not able to find cached extensions corresponding to {module_name}"
        )
        print("Nothing to do! ✨ 🍰 ✨.")
        return

    if path_pythran_dir_jit.exists() and query_yes_no(
        f"Do you confirm that you want to delete the cached files for {module_name}",
        default="y",
        force=force,
    ):
        print(f"Remove directory {path_pythran_dir_jit}")
        shutil.rmtree(fspath(path_pythran_dir_jit))

    if path_pythran.exists() or path_ext.exists():
        if query_yes_no(
            f"Do you confirm that you want to delete the AOT cache for {module_name}",
            default="y",
            force=force,
        ):
            for path in (path_pythran, path_ext):
                if path.exists():
                    path.unlink()


def is_method(func):
    """Determine wether a function is going to be used as a method"""
    signature = inspect.signature(func)
    try:
        answer = next(iter(signature.parameters.keys())) == "self"
    except StopIteration:
        answer = False

    return answer



"""Configuration
================

Transonic is sensible to the environment variables:

- :code:`TRANSONIC_DIR` can be set to control where the cached files are
  saved.

- :code:`TRANSONIC_COMPILE_AT_IMPORT` can be set to enable a mode for which
  Transonic compiles at import time the Pythran file associated with the
  imported module. This behavior can also be triggered programmatically by using
  the function :code:`set_compile_at_import`.

- :code:`TRANSONIC_NO_REPLACE` can be set to disable all code replacements.
  This is useful only when measuring code coverage.

- :code:`FLUID_COMPILE_JIT` can be set to false to disable the
  compilation of jited functions. This can be useful for unittests.

Bye the way, for performance, it is important to configure Pythran with a file
`~/.pythranrc
<https://pythran.readthedocs.io/en/latest/MANUAL.html#customizing-your-pythranrc>`_:

For Linux, I use:

.. code::

    [pythran]
    complex_hook = True

    [compiler]
    CXX=clang++-6.0
    CC=clang-6.0
    blas=openblas

"""

import os
from pathlib import Path
from distutils.util import strtobool

path_root = Path(os.environ.get("TRANSONIC_DIR", Path.home() / ".transonic"))

has_to_replace = not strtobool(os.environ.get("TRANSONIC_NO_REPLACE", "0"))



import sys

import time

import pytest

from . import path_data_tests
from .run import run
from . import util
from .compat import rmtree
from .mpi import nb_proc

path_dir_out = path_data_tests / "__pythran__"


@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_pythran_files():
    if path_dir_out.exists():
        rmtree(path_dir_out)

    sys.argv = f"transonic -np {path_data_tests / '*.py'}".split()
    run()

    sys.argv = f"transonic -np {path_data_tests}".split()
    run()

    paths = path_data_tests.glob("*.py")
    sys.argv = ["transonic", "-np"] + [str(path) for path in paths]
    run()


@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_pythran_simple():

    sys.argv = "transonic --version".split()
    run()

    sys.argv = ["transonic"]
    run()


@pytest.mark.skipif(nb_proc > 1, reason="No commandline in MPI")
def test_create_trans_classic():

    util.input = lambda: "y"

    if path_dir_out.exists():
        rmtree(path_dir_out)

    path_file = path_data_tests / "classic.py"
    sys.argv = f"transonic -np {path_file}".split()
    run()

    print("after first build")
    run()

    time.sleep(0.02)
    path_file.touch()
    print("after touch")

    run()

    path_file_pythran = path_data_tests / "__pythran__/classic.py"
    path_file_pythran.unlink()

    print("after unlink")
    run()

    sys.argv = f"transonic -cc {path_file}".split()
    run()

    sys.argv = f"transonic {path_file}".split()
    run()

    sys.argv = f"transonic -np {path_file_pythran}".split()
    run()



import numpy as np

# transonic import numpy as np

from transonic import jit, include, boost
from .mpi import Path


@include(used_by="func1")
def func0(a):
    return 2 * a


@include(used_by="func1")
def func():
    return 1


@jit
def func1(a: "int[][] or float[]", l: "int list"):
    tmp = np.exp(sum(l))
    result = tmp * a * func0(a) + func()
    return result


@jit()
def func2(a):
    return a


# weird but done on purpose for a better coverage
Path(__file__).touch()


@jit()
def func2(a):
    return a


@jit()
def func_dict(d: "str: float dict"):
    return d.popitem()


@boost
class MyClass:
    def __init__(self):
        self.attr0 = self.attr1 = 1

    @jit
    def myfunc(self, arg):
        return self.attr1 + self.attr0 + np.abs(arg)

    def check(self):
        assert self.myfunc(1) == 3



"""
See https://bitbucket.org/fluiddyn/fluidpythran/issues/7/

We can use the decorator transonic.include but here we don't use it bacause
it should not be difficult to detect that `func` uses `mylog`.

"""
import numpy as np

from transonic import boost

# transonic import numpy as np


def mylog(arr):
    return np.log(arr)


@boost
def func(a: float, b: float):
    return (a * mylog(b)).max()




from functools import wraps

import numpy as np

# pythran export __method__Transmitter____call__(float, float[])


def __for_method__Transmitter____call__(self_freq, inp):
    return inp * np.exp(np.arange(len(inp))*self_freq*1j)


def __method__Transmitter____call__(self, inp):
    self_freq = self.freq
    return __for_method__Transmitter____call__(self_freq, inp)


def __call__(self, inp: "float[]"):
    """My docstring"""
    return inp * np.exp(np.arange(len(inp))*self.freq*1j)


class Transmitter():

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    __call__ = wraps(__call__)(__method__Transmitter____call__)


from . import util
from .util import query_yes_no


def test_query_yes_no():

    util.input = lambda: "y"

    query_yes_no("test", default="y")
    query_yes_no("test", default="n")
    query_yes_no("test")

    util.input = lambda: ""
    query_yes_no("test", default="y")

    query_yes_no("test", force=True)



from functools import wraps

import numpy as np

import pythran_manual


class FluidPythranTemporaryMethod:
    def __init__(self, func):
        self.func = func

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError("Do not call this function!")


def trans_def_method(func):
    return FluidPythranTemporaryMethod(func)


def boost(cls):

    cls_name = cls.__name__

    for key, value in cls.__dict__.items():
        if not isinstance(value, FluidPythranTemporaryMethod):
            continue
        func = value.func
        func_name = func.__name__

        name_pythran_func = f"__for_method__{cls_name}__{func_name}"
        pythran_func = pythran_manual.__dict__[name_pythran_func]

        name_var_code_new_method = f"__code_new_method__{cls_name}__{func_name}"
        code_new_method = pythran_manual.__dict__[name_var_code_new_method]

        namespace = {"pythran_func": pythran_func}
        exec(code_new_method, namespace)
        setattr(cls, key, wraps(func)(namespace["new_method"]))

    return cls


@boost
class Transmitter:

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    @trans_def_method
    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)


if __name__ == "__main__":
    inp = np.ones(2)
    trans = Transmitter(1.)
    trans(inp)



import beniget
import gast as ast


with open("issue6_import.py") as file:
    code = file.read()

module = ast.parse(code)

duc = beniget.DefUseChains()
duc.visit(module)

ancestors = beniget.Ancestors()
ancestors.visit(module)

boost = [d for d in duc.locals[module] if d.name() == "boost"][0]

for use in boost.uses:
    # we're interested in the parent of the decorator
    parents = ancestors.parents[use.node]
    # direct parent of the decorator is the function
    fdef = parents[-1]
    print(fdef.name)

chain = duc.chains[fdef]


class Capture(ast.NodeVisitor):
    def __init__(self, module_node):
        # initialize def-use chains
        self.duc = beniget.DefUseChains()
        self.duc.visit(module_node)
        self.uses_local_def = set()  # uses of local definitions
        self.captured = set()  # identifiers that don't belong to local uses

    def visit_FunctionDef(self, node):
        # initialize the set of node using a local variable
        for def_ in self.duc.locals[node]:
            print("def_.name()", def_.name())
            print(list((use.name(), use.node.ctx) for use in def_.uses))
            self.uses_local_def.update(use.node for use in def_.uses)

        self.generic_visit(node)

    def visit_Name(self, node):
        # register load of identifiers not locally defined
        print("node.id", node.id, node.ctx)
        if isinstance(node.ctx, ast.Load):
            if node not in self.uses_local_def:
                self.captured.add(node)


capture = Capture(module)

for node in fdef.decorator_list:
    if node.id == "boost":
        fdef.decorator_list.remove(node)

capture.visit(fdef)

print(list(node.id for node in capture.captured))

node = next(iter(capture.captured))



from functools import wraps

import numpy as np

import pythran_manual

code = """

def new_method(self, inp):
    return pythran_func(self.freq, inp)

"""


def my_decor(func):
    func_name = func.__name__

    name_pythran_func = "__for_method__Transmitter__" + func_name

    pythran_func = pythran_manual.__dict__[name_pythran_func]

    namespace = {"pythran_func": pythran_func}
    exec(code, namespace)
    return wraps(func)(namespace["new_method"])


class Transmitter:

    freq: float

    def __init__(self, freq):
        self.freq = float(freq)

    @my_decor
    def __call__(self, inp: "float[]"):
        """My docstring"""
        return inp * np.exp(np.arange(len(inp)) * self.freq * 1j)



import importlib

from . import Transonic, mpi

from .compat import rmtree
from .mpi import Path


def test_not_transonified():

    path_for_test = Path(__file__).parent / "for_test_init.py"

    path_output = path_for_test.parent / "__pythran__"

    if path_output.exists() and mpi.rank == 0:
        rmtree(path_output)
    mpi.barrier()

    from . import for_test_init

    importlib.reload(for_test_init)

    from .for_test_init import func, func1, check_class

    func(1, 3.14)
    func1(1.1, 2.2)
    check_class()


def test_use_pythran_false():
    Transonic(use_transonified=False)



from setuptools import setup, find_packages
from runpy import run_path as run_path_
from pathlib import Path
import platform

if platform.python_implementation() == "PyPy":

    def run_path(path):
        return run_path_(str(path))


else:
    run_path = run_path_

here = Path(__file__).parent.absolute()

d = run_path(here / "transonic/_version.py")
__version__ = d["__version__"]
__about__ = d["__about__"]

print(__about__)

path_data_tests = here / "data_tests"
if path_data_tests.exists():
    with open("transonic/_path_data_tests.py", "w") as file:
        file.write(
            "from pathlib import Path\n\n"
            f'path_data_tests = Path("{path_data_tests}")\n'
        )

setup(
    version=__version__,
    packages=find_packages(exclude=["doc"]),
    entry_points={"console_scripts": ["transonic = transonic.run:run"]},
)



import sys
import os
from time import sleep

import numpy as np

try:
    import pythran
except ImportError:
    pythran = None

from .util import path_jit_classes
from .justintime import path_jit, modules
from .pythranizer import scheduler, wait_for_all_extensions
from . import mpi
from .compat import rmtree

scheduler.nb_cpus = 2

module_name = "transonic.for_test_justintime"

str_relative_path = module_name.replace(".", os.path.sep)

path_jit = mpi.PathSeq(path_jit)

path_pythran_dir = path_jit / str_relative_path
path_classes_dir = path_jit_classes / str_relative_path
path_classes_dir1 = path_jit / path_jit_classes.name / str_relative_path


def delete_pythran_files(func_name):
    for path_pythran_file in path_pythran_dir.glob(func_name + "*"):
        if path_pythran_file.exists():
            path_pythran_file.unlink()


if mpi.rank == 0:
    delete_pythran_files("func1")
    delete_pythran_files("func2")
    delete_pythran_files("func_dict")

    if path_classes_dir.exists():
        rmtree(path_classes_dir)
    if path_classes_dir1.exists():
        rmtree(path_classes_dir1)

mpi.barrier()


def test_jit():

    from .for_test_justintime import func1

    a = np.arange(2)
    b = [1, 2]

    for _ in range(2):
        func1(a, b)
        sleep(0.1)


def test_jit_simple():

    from .for_test_justintime import func2

    func2(1)

    if not pythran:
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func2"]

    for _ in range(100):
        func2(1)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func2(1)
            break

    del sys.modules[module_name]
    del modules[module_name]

    from .for_test_justintime import func2

    func2(1)


def test_jit_dict():
    from .for_test_justintime import func_dict

    d = dict(a=1, b=2)
    func_dict(d)

    if not pythran:
        return

    mod = modules[module_name]
    cjit = mod.jit_functions["func_dict"]

    d = dict(a=1, b=2)
    func_dict(d)

    wait_for_all_extensions()

    for _ in range(100):
        d = dict(a=1, b=2)
        func_dict(d)
        sleep(0.1)
        if not cjit.compiling:
            sleep(0.1)
            func_dict(d)
            break


def test_jit_method():
    from .for_test_justintime import MyClass

    obj = MyClass()
    obj.check()

    if not pythran:
        return

    obj = MyClass()
    obj.check()

    wait_for_all_extensions()

    obj.check()



from distutils.core import setup, Extension
from pathlib import Path
import numpy as np
from transonic.dist import make_backend_files, init_pythran_extensions, ParallelBuildExt
from transonic.compat import fspath


path_sources = Path(__file__).parent.absolute()
include_dirs = [np.get_include()]

try:
    from Cython.Build import cythonize
except ImportError:
    # from setuptools import Extension
    extensions = [
        Extension(
            "add_cython",
            include_dirs=[fspath(path_sources)] + include_dirs,
            libraries=["m"],
            library_dirs=[],
            sources=[fspath(path_sources / "add_cython.c")],
        )
    ]
    print(extensions)
else:
    extensions = cythonize(
        Extension("add_cython", ["add_cython.pyx"], include_dirs=include_dirs)
    )

make_backend_files([path_sources / "add.py"])
extensions.extend(init_pythran_extensions(".", include_dirs=include_dirs))

setup(
    name="add",
    ext_modules=extensions,
    script_name="setup.py",
    script_args=["build_ext"],
    cmdclass=dict(build_ext=ParallelBuildExt),
)



"""
See https://bitbucket.org/fluiddyn/fluidpythran/issues/6/

Here, the function uses np (numpy), but it should of course work for all
packages.

Related question: do we support usage of global variables in "boosted"
functions?

"""
import numpy as np

from transonic import boost

my_constant = 3


@boost
def func(a: float, b: float):

    c = a + b
    return (a * np.log(c)).max() * my_constant


if __name__ == "__main__":
    result1 = func(1., 1.)
    result2 = func(2., 2.)


"""User runtime API for the ahead-of-time compilation
=====================================================

User API
--------

.. autofunction:: boost

.. autofunction:: make_signature

.. autoclass:: Transonic
   :members:
   :private-members:

Internal API
------------

.. autofunction:: _get_transonic_calling_module

.. autoclass:: CheckCompiling
   :members:
   :private-members:

"""

import inspect
import time
import subprocess
import os
import functools
import sys

from .util import (
    get_module_name,
    has_to_compile_at_import,
    import_from_path,
    has_to_build,
    modification_date,
    is_method,
    path_jit_classes,
)

from .pythranizer import compile_extension, name_ext_from_path_backend, ext_suffix

from .annotation import (
    make_signature_from_template_variables,
    make_signatures_from_typehinted_func,
)

from .transpiler import produce_code_class

from .log import logger
from . import mpi
from .mpi import Path
from .config import has_to_replace
from .compat import fspath

if mpi.nb_proc == 1:
    mpi.has_to_build = has_to_build
    mpi.modification_date = modification_date

is_transpiling = False
modules = {}

path_jit_classes = mpi.Path(path_jit_classes)


def _get_transonic_calling_module(index_frame: int = 2):
    """Get the Transonic instance corresponding to the calling module

    Parameters
    ----------

    index_frame : int

      Index (in :code:`inspect.stack()`) of the frame to be selected

    """

    try:
        frame = inspect.stack()[index_frame]
    except IndexError:
        print("index_frame", index_frame)
        print([frame[1] for frame in inspect.stack()])
        raise

    module_name = get_module_name(frame)

    if module_name in modules:
        ts = modules[module_name]
        if (
            ts.is_transpiling != is_transpiling
            or ts._compile_at_import_at_creation != has_to_compile_at_import()
            or hasattr(ts, "path_mod")
            and ts.path_mod.exists()
            and mpi.has_to_build(ts.path_pythran, ts.path_mod)
        ):
            ts = Transonic(frame=frame, reuse=False)
    else:
        ts = Transonic(frame=frame, reuse=False)

    return ts


def boost(obj):
    """Decorator to declare that an object can "use" pythran

    Parameters
    ----------

    obj: a function, a method or a class

    """

    ts = _get_transonic_calling_module()
    return ts.boost(obj)


def make_signature(func, **kwargs):
    """Create signature for a function with values for the template types

    Parameters
    ----------

    func: a function

    kwargs : dict

      The template types and their value

    """
    if not is_transpiling:
        return

    ts = _get_transonic_calling_module()
    ts.make_signature(func, **kwargs)


class CheckCompiling:
    """Check if the module is being compiled and replace the module and the function"""

    def __init__(self, ts, func):
        self.has_been_replaced = False
        self.ts = ts
        self.func = func

    def __call__(self, *args, **kwargs):

        if self.has_been_replaced:
            return self.func(*args, **kwargs)

        ts = self.ts
        if ts.is_compiling and not ts.process.is_alive():
            ts.is_compiling = False
            time.sleep(0.1)
            ts.module_pythran = import_from_path(
                ts.path_extension, ts.module_pythran.__name__
            )
            assert hasattr(self.ts.module_pythran, "__pythran__")
            ts.is_compiled = True

        if not ts.is_compiling:
            self.func = getattr(ts.module_pythran, self.func.__name__)
            self.has_been_replaced = True

        return self.func(*args, **kwargs)


class Transonic:
    """
    Representation of a module using ahead-of-time transonic commands

    Parameters
    ----------

    use_transonified : bool (optional, default True)

      If False, don't use the pythranized versions at run time

    frame : int (optional)

      (Internal) Index (in :code:`inspect.stack()`) of the frame to be selected

    reuse : bool (optional, default True)

      (Internal) If True, do not recreate an instance.

    """

    def __init__(self, use_transonified=True, frame=None, reuse=True):

        if frame is None:
            frame = inspect.stack()[1]

        self.module_name = module_name = get_module_name(frame)

        self._compile_at_import_at_creation = has_to_compile_at_import()

        if reuse and module_name in modules:
            ts = modules[module_name]
            for key, value in ts.__dict__.items():
                self.__dict__[key] = value
            return

        self.names_template_variables = {}

        self.is_transpiling = is_transpiling
        self.has_to_replace = has_to_replace

        if is_transpiling:
            self.functions = {}
            self.classes = {}
            self.signatures_func = {}
            self.included_functions = []
            modules[module_name] = self
            self.is_transpiled = False
            self.is_compiled = False
            return

        self.is_compiling = False

        if not use_transonified or not has_to_replace:
            self.is_transpiled = False
            self.is_compiled = False
            return

        if "." in module_name:
            package, module_short_name = module_name.rsplit(".", 1)
            module_pythran_name = package + "."
        else:
            module_short_name = module_name
            module_pythran_name = ""

        module_pythran_name += "__pythran__." + module_short_name

        self.path_mod = path_mod = Path(frame.filename)
        self.path_pythran = path_pythran = (
            path_mod.parent / "__pythran__" / (module_short_name + ".py")
        )

        path_ext = None

        if has_to_compile_at_import() and path_mod.exists():
            if mpi.has_to_build(path_pythran, path_mod):
                if path_pythran.exists():
                    time_pythran = mpi.modification_date(path_pythran)
                else:
                    time_pythran = 0

                returncode = None
                if mpi.rank == 0:
                    print(f"Running transonic on file {path_mod}... ", end="")
                    # better to do this in another process because the file is already run...
                    os.environ["TRANSONIC_NO_MPI"] = "1"
                    returncode = subprocess.call(
                        [
                            sys.executable,
                            "-m",
                            "transonic.run",
                            "-np",
                            fspath(path_mod),
                        ]
                    )
                    del os.environ["TRANSONIC_NO_MPI"]
                returncode = mpi.bcast(returncode)

                if returncode != 0:
                    raise RuntimeError(
                        "transonic does not manage to produce the Pythran "
                        f"file for {path_mod}"
                    )

                if mpi.rank == 0:
                    print("Done!")

                path_ext = path_pythran.with_name(
                    name_ext_from_path_backend(path_pythran)
                )

                time_pythran_after = mpi.modification_date(path_pythran)
                # We have to touch the files to signal that they are up-to-date
                if time_pythran_after == time_pythran and mpi.rank == 0:
                    if not has_to_build(path_ext, path_pythran):
                        path_pythran.touch()
                        if path_ext.exists():
                            path_ext.touch()
                    else:
                        path_pythran.touch()

        path_ext = path_ext or path_pythran.with_name(
            name_ext_from_path_backend(path_pythran)
        )

        self.path_extension = path_ext

        if (
            has_to_compile_at_import()
            and path_mod.exists()
            and not self.path_extension.exists()
        ):
            if mpi.rank == 0:
                print("Launching Pythran to compile a new extension...")
            self.process = compile_extension(
                path_pythran, name_ext_file=self.path_extension.name
            )
            self.is_compiling = True
            self.is_compiled = False

        self.is_transpiled = True

        if not path_ext.exists() and not self.is_compiling:
            path_ext_alt = path_pythran.with_suffix(ext_suffix)
            if path_ext_alt.exists():
                self.path_extension = path_ext = path_ext_alt

        self.reload_module_pythran(module_pythran_name)

        if self.is_transpiled:
            self.is_compiled = hasattr(self.module_pythran, "__pythran__")
            if self.is_compiled:
                module = inspect.getmodule(frame[0])
                # module can be None if (at least) it has been run with runpy
                if module is not None:
                    module.__pythran__ = self.module_pythran.__pythran__
                    module.__transonic__ = self.module_pythran.__transonic__

            if hasattr(self.module_pythran, "arguments_blocks"):
                self.arguments_blocks = getattr(
                    self.module_pythran, "arguments_blocks"
                )

        modules[module_name] = self

    def reload_module_pythran(self, module_pythran_name=None):
        if module_pythran_name is None:
            module_pythran_name = self.module_pythran.__name__
        if self.path_extension.exists() and not self.is_compiling:
            self.module_pythran = import_from_path(
                self.path_extension, module_pythran_name
            )
        elif self.path_pythran.exists():
            self.module_pythran = import_from_path(
                self.path_pythran, module_pythran_name
            )
        else:
            self.is_transpiled = False
            self.is_compiled = False

    def transonic_def(self, func):
        """Decorator used for functions

        Parameters
        ----------

        func: a function

        """
        if is_method(func):
            return self.transonic_def_method(func)

        if is_transpiling:
            self.functions[func.__name__] = func
            return func

        if not has_to_replace or not self.is_transpiled:
            return func

        if not hasattr(self.module_pythran, func.__name__):
            self.reload_module_pythran()

        try:
            func_tmp = getattr(self.module_pythran, func.__name__)
        except AttributeError:
            logger.warning("Pythran file does not seem to be up-to-date.")
            func_tmp = func

        if self.is_compiling:
            return functools.wraps(func)(CheckCompiling(self, func_tmp))

        return func_tmp

    def transonic_def_method(self, func):
        """Decorator used for methods

        Parameters
        ----------

        func: a function

        """
        if is_transpiling:
            func.__transonic__ = "trans_def_method"
            return func

        if not has_to_replace or not self.is_transpiled:
            return func

        return TransonicTemporaryMethod(func)

    def boost(self, obj):
        """Universal decorator for AOT compilation

        Used for functions, methods and classes.
        """
        if isinstance(obj, type):
            return self.transonic_class(obj)
        else:
            return self.transonic_def(obj)

    def transonic_class(self, cls: type):
        """Decorator used for classes

        Parameters
        ----------

        cls: a class

        """
        if is_transpiling:
            self.classes[cls.__name__] = cls
            return cls

        jit_methods = {
            key: value
            for key, value in cls.__dict__.items()
            if isinstance(value, TransonicTemporaryJITMethod)
        }

        if jit_methods:
            cls = jit_class(cls, jit_methods)

        if not has_to_replace or not self.is_transpiled:
            return cls

        cls_name = cls.__name__

        for key, value in cls.__dict__.items():
            if not isinstance(value, TransonicTemporaryMethod):
                continue
            func = value.func
            func_name = func.__name__

            name_pythran_func = f"__for_method__{cls_name}__{func_name}"
            name_var_code_new_method = (
                f"__code_new_method__{cls_name}__{func_name}"
            )

            if not hasattr(self.module_pythran, name_pythran_func):
                self.reload_module_pythran()

            try:
                pythran_func = getattr(self.module_pythran, name_pythran_func)
                code_new_method = getattr(
                    self.module_pythran, name_var_code_new_method
                )
            except AttributeError:
                logger.warning("Pythran file does not seem to be up-to-date.")
                # setattr(cls, key, func)
            else:
                namespace = {"pythran_func": pythran_func}
                exec(code_new_method, namespace)
                setattr(cls, key, functools.wraps(func)(namespace["new_method"]))

        return cls

    def make_signature(self, func, _signature=None, **kwargs):
        """Create signature for a function with values for the template types

        Parameters
        ----------

        func: a function

        kwargs : dict

          The template types and their value

        """
        signatures = make_signature_from_template_variables(
            func, _signature=_signature, **kwargs
        )

        if func.__name__ not in self.signatures_func:
            self.signatures_func[func.__name__] = []

        self.signatures_func[func.__name__].extend(signatures)

    def use_block(self, name):
        """Use the pythranized version of a code block

        Parameters
        ----------

        name : str

          The name of the block.

        """
        if not self.is_transpiled:
            raise ValueError(
                "`use_block` has to be used protected by `if ts.is_transpiled`"
            )

        if self.is_compiling and not self.process.is_alive():
            self.is_compiling = False
            time.sleep(0.1)
            self.module_pythran = import_from_path(
                self.path_extension, self.module_pythran.__name__
            )
            assert hasattr(self.module_pythran, "__pythran__")
            self.is_compiled = True

        func = getattr(self.module_pythran, name)
        argument_names = self.arguments_blocks[name]

        frame = inspect.currentframe()
        try:
            locals_caller = frame.f_back.f_locals
        finally:
            del frame

        arguments = [locals_caller[name] for name in argument_names]
        return func(*arguments)

    def _make_signatures_from_annotations(self):
        """Make the signatures from annotations if it is possible

        Useful when there are only "not templated" types.

        """
        for func in self.functions.values():
            signatures = make_signatures_from_typehinted_func(func)
            if func.__name__ not in self.signatures_func:
                self.signatures_func[func.__name__] = []
            self.signatures_func[func.__name__].extend(signatures)

    def include(self, func):
        self.included_functions.append(func)


class TransonicTemporaryMethod:
    """Internal temporary class for methods"""

    def __init__(self, func):
        self.func = func

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError(
            "Did you forget to decorate a class using methods decorated "
            "with transonic? Please decorate it with @boost."
        )


def include(func=None, used_by=None):
    """Decorator to record that the function has to be included in a Pythran file

    """
    if func is not None and used_by is None:
        if not callable(func):
            func, used_by = used_by, func

    decor = Include(used_by=used_by)
    if callable(func):
        decor._decorator_no_arg = True
        return decor(func)
    else:
        return decor


class Include:
    """Decorator used internally by the public include decorator

    @include

    @include(used_by=("func0", "func1")

    """

    def __init__(self, used_by=None):
        self.used_by = used_by
        self._decorator_no_arg = False

    def __call__(self, func):

        if self._decorator_no_arg:
            index_frame = 3
        else:
            index_frame = 2

        if self.used_by is not None:
            from .justintime import _get_module_jit

            jit_mod = _get_module_jit(index_frame)
            jit_mod.record_used_function(func, self.used_by)

        if not is_transpiling:
            return func

        ts = _get_transonic_calling_module(index_frame)
        ts.include(func)
        return func


class TransonicTemporaryJITMethod:
    """Internal temporary class for JIT methods"""

    __transonic__ = "jit_method"

    def __init__(self, func, native, xsimd, openmp):
        self.func = func
        self.native = native
        self.xsimd = xsimd
        self.openmp = openmp

    def __call__(self, self_bis, *args, **kwargs):
        raise RuntimeError(
            "Did you forget to decorate a class using methods decorated "
            "with transonic? Please decorate it with @boost."
        )


def jit_class(cls, jit_methods):
    """Modify the class by replacing jit methods

    1. create a Python file with @jit functions and methods
    2. import the file
    3. replace the methods

    """
    if not has_to_replace:
        return cls

    cls_name = cls.__name__
    mod_name = cls.__module__

    module = sys.modules[mod_name]

    # 1. create a Python file with @jit functions and methods
    python_path_dir = path_jit_classes / mod_name.replace(".", os.path.sep)
    python_path = python_path_dir / (cls_name + ".py")

    if mpi.has_to_build(python_path, module.__file__):
        if mpi.rank == 0:
            python_path = mpi.PathSeq(python_path)

            # quick fix to keep the "#transonic import" commands
            transonic_imports = []
            with open(module.__file__) as file:
                for line in file:
                    if line.startswith("# transonic ") and " import " in line:
                        transonic_imports.append(line)
                        transonic_imports.append(line[12:])

            python_code = "\n".join(transonic_imports) + "\n"

            python_code += produce_code_class(cls, jit=True)

            has_to_write = None
            if python_path.exists():
                with open(python_path, "r") as file:
                    python_code_file = file.read()

                if python_code_file != python_code:
                    has_to_write = True
            else:
                has_to_write = True

            if has_to_write:
                python_path_dir.mkdir(exist_ok=True, parents=True)
                with open(python_path, "w") as file:
                    file.write(python_code)
            python_path = mpi.Path(python_path)
        mpi.barrier()

    # 2. import the file
    python_mod_name = path_jit_classes.name + "." + mod_name + "." + cls_name
    module = import_from_path(python_path, python_mod_name)

    # 3. replace the methods
    for name_method, method in jit_methods.items():
        func = method.func
        name_new_method = f"__new_method__{cls.__name__}__{name_method}"
        new_method = getattr(module, name_new_method)
        setattr(cls, name_method, functools.wraps(func)(new_method))

    return cls



from transonic import boost, jit


@boost
class OtherClass:
    """Note that there is no type annotations at all!

    The Pythran signature is created at run time with the types of the
    attributes and the arguments.

    """

    def __init__(self, arg):
        self.attr0 = self.attr1 = 2 * arg

    @jit
    def calcul(self, a, b):
        return a * self.attr0 + b * self.attr1
